@@ -27,6 +27,11 @@ from chroniker.models import Job, Log, JobDependency, Monitor
 from chroniker.utils import get_admin_changelist_url
 from chroniker.widgets import ImproveRawIdFieldsFormTabularInline
 
+try:
+    from admin_steroids.queryset import ApproxCountQuerySet
+except ImportError:
+    ApproxCountQuerySet = None
+
 class HTMLWidget(forms.Widget):
     def __init__(self,rel=None, attrs=None):
         self.rel = rel
@@ -73,6 +78,7 @@ class JobAdmin(admin.ModelAdmin):
         'disable_jobs',
         'run_selected_jobs',
         'toggle_enabled',
+        'clear_stalled',
     )
     list_display = (
         'name',
@@ -115,12 +121,15 @@ class JobAdmin(admin.ModelAdmin):
         'estimated_completion_datetime_str',
         'monitor_records',
         'current_hostname',
+        'current_pid',
         'job_type',
+        'is_due',
     )
     list_display_links = ('name', )
     list_filter = (
         'frequency',
         'enabled',
+        'is_running',
         'hostname',
         'is_monitor',
     )
@@ -134,22 +143,45 @@ class JobAdmin(admin.ModelAdmin):
                 'args',
                 'hostname',
                 'current_hostname',
-                'enabled',
-                #'check_is_running',
+                'current_pid',
+            )
+        }),
+        ('Status', {
+            'classes': ('wide',),
+            'fields': (
                 'is_running',
-                'force_run',
-                'force_stop',
-                'view_logs_button',
+                'is_due',
+                'is_fresh', 
                 'last_run_successful',
-                'is_fresh',
-                'last_heartbeat',
-                'last_run_start_timestamp',
-                'last_run',
                 'total_parts',
                 'total_parts_complete',
                 'progress_percent_str',
                 'estimated_completion_datetime_str',
+                'last_heartbeat',
+                'last_run_start_timestamp',
+                'last_run',
+            )
+        }),
+        ('Flags', {
+            'classes': ('wide',),
+            'fields': (
+                'enabled',
+                'force_run',
+                'force_stop',
+            )
+        }),
+        ('Logging', {
+            'classes': ('wide',),
+            'fields': (
+                'view_logs_button',
+                'log_stdout',
+                'log_stderr',
                 'maximum_log_entries',
+            )
+        }),
+        ('Monitor', {
+            'classes': ('wide',),
+            'fields': (
                 'is_monitor',
                 'monitor_url',
                 'monitor_error_template',
@@ -167,24 +199,36 @@ class JobAdmin(admin.ModelAdmin):
         }),
         ('Frequency options', {
             'classes': ('wide',),
-            'fields': ('frequency', 'next_run', 'params',)
+            'fields': (
+                'frequency',
+                'next_run',
+                'params',
+                'timeout_seconds',
+            )
         }),
     )
     search_fields = (
         'name',
         'hostname',
         'current_hostname',
+        'current_pid',
     )
     
     inlines = (
        JobDependencyInline,
     )
     
+    def queryset(self, *args, **kwargs):
+        qs = super(JobAdmin, self).queryset(*args, **kwargs)
+        if ApproxCountQuerySet is not None:
+            qs = qs._clone(klass=ApproxCountQuerySet)
+        return qs
+    
     def get_readonly_fields(self, request, obj=None):
         fields = list(self.readonly_fields)
         # Allow manual clearing of is_running if the cron job has become stuck.
-        if obj and obj.is_fresh():
-            fields.append('is_running')
+#        if obj and obj.is_fresh():
+#            fields.append('is_running')
         return fields
     
     def last_run_with_link(self, obj=None):
@@ -216,6 +260,7 @@ class JobAdmin(admin.ModelAdmin):
         return not obj.check_is_running()
     check_is_complete.short_description = _('is complete')
     check_is_complete.boolean = True
+    check_is_complete.admin_order_field = 'is_running'
     
     def get_timeuntil(self, obj=None):
         if not obj or not obj.id or not obj.next_run:
@@ -242,7 +287,7 @@ class JobAdmin(admin.ModelAdmin):
         if not obj or not obj.id:
             return ''
         url = '%d/run/?inline=1' % obj.id
-        return '<a href="%s"><input type="button" value="Run" /></a>' % url
+        return '<a href="%s" class="button">Run</a>' % url
     run_button.allow_tags = True
     run_button.short_description = 'Run'
     
@@ -253,8 +298,7 @@ class JobAdmin(admin.ModelAdmin):
         vars = dict(url=url, disabled='')
         if not obj.is_running:
             vars['disabled'] = 'disabled'
-        s = ('<a href="%(url)s"><input type="button" %(disabled)s ' + \
-            'value="Stop" /></a>') % vars
+        s = ('<a href="%(url)s" class="button" %(disabled)s>Stop</a>') % vars
         return s
     stop_button.allow_tags = True
     stop_button.short_description = 'Stop'
@@ -264,51 +308,41 @@ class JobAdmin(admin.ModelAdmin):
             return ''
         q = obj.logs.all()
         url = get_admin_changelist_url(Log)
-        return ('<a href="%s?job=%d" target="_blank">'
-            '<input type="button" value="View %i" /></a>') % \
+        return ('<a href="%s?job=%d" target="_blank" class="button">View&nbsp;%i</a>') % \
             (url, obj.id, q.count())
     view_logs_button.allow_tags = True
     view_logs_button.short_description = 'Logs'
     
-    def run_job_view(self, request, pk):
+    def run_job_view(self, request, job_id):
         """
         Runs the specified job.
         """
-        try:
-            job = Job.objects.get(pk=pk)
-        except Job.DoesNotExist:
-            raise Http404
-        # Rather than actually running the Job right now, we
-        # simply force the Job to be run by the next cron job
-        job.force_run = True
-        job.save()
+        Job.objects.filter(id=job_id).update(
+            force_run = True,
+            force_stop = False,
+        )
         self.message_user(
             request,
-            _('The job "%(job)s" has been scheduled to run.') \
-                % {'job': job})
+            _('Job %(job)s has been signalled to start running immediately.') \
+                % {'job': job_id})
         if 'inline' in request.GET:
             redirect = request.path + '../../'
         else:
             redirect = request.REQUEST.get('next', request.path + "../")
         return HttpResponseRedirect(redirect)
     
-    def stop_job_view(self, request, pk):
+    def stop_job_view(self, request, job_id):
         """
         Stop the specified job.
         """
-        try:
-            job = Job.objects.get(pk=pk)
-        except Job.DoesNotExist:
-            raise Http404
-        # Rather than actually running the Job right now, we
-        # simply force the Job to be run by the next cron job
-        job.force_run = False
-        job.force_stop = True
-        job.save()
+        Job.objects.filter(id=job_id).update(
+            force_run = False,
+            force_stop = True,
+        )
         self.message_user(
             request,
-            _('The job "%(job)s" is being signalled to stop.') \
-                % {'job': job})
+            _('Job %(job)s has been signalled to stop running immediately.') \
+                % {'job': job_id})
         if 'inline' in request.GET:
             redirect = request.path + '../../'
         else:
@@ -337,6 +371,16 @@ class JobAdmin(admin.ModelAdmin):
         self.message_user(request, "%s successfully set to run." % message_bit)
     run_selected_jobs.short_description = "Force run selected jobs"
     
+    def clear_stalled(self, request, queryset):
+        reset_count = 0
+        for job in queryset:
+            if not job.is_fresh():
+                reset_count += 1
+                job.is_running = False
+                job.save()
+        self.message_user(request, 'Cleared %i stalled jobs.' % (reset_count,))
+    clear_stalled.short_description = 'Mark the selected stalled %(verbose_name_plural)s as not running'
+        
     def toggle_enabled(self, request, queryset):
         for row in queryset:
             row.enabled = not row.enabled
@@ -350,9 +394,7 @@ class JobAdmin(admin.ModelAdmin):
     toggle_enabled.short_description = "Toggle enabled flag on selected jobs"
     
     def disable_jobs(self, request, queryset):
-        for row in queryset:
-            row.enabled = False
-            row.save()
+        queryset.update(enabled=False)
         rows_updated = queryset.count()
         if rows_updated == 1:
             message_bit = "1 job was toggled"
@@ -362,9 +404,7 @@ class JobAdmin(admin.ModelAdmin):
     disable_jobs.short_description = "Disable selected jobs"
     
     def enable_jobs(self, request, queryset):
-        for row in queryset:
-            row.enabled = True
-            row.save()
+        queryset.update(enabled=True)
         rows_updated = queryset.count()
         if rows_updated == 1:
             message_bit = "1 job was toggled"
@@ -397,6 +437,7 @@ class JobAdmin(admin.ModelAdmin):
         return super(JobAdmin, self).formfield_for_dbfield(db_field, **kwargs)
 
 class LogAdmin(admin.ModelAdmin):
+    
     list_display = (
         'job_name',
         'run_start_datetime',
@@ -404,15 +445,24 @@ class LogAdmin(admin.ModelAdmin):
         'duration_seconds',
         'duration_str',
         'job_success',
+        'on_time',
+        'hostname',
         #'stdout_sample',
         #'stderr_sample',
     )
+    
+    list_filter = (
+        'success',
+        'on_time',
+    )
+    
     search_fields = (
         'stdout',
         'stderr',
         'job__name',
         'job__command',
     )
+    
     readonly_fields = (
         'run_start_datetime',
         'run_end_datetime',
@@ -422,6 +472,7 @@ class LogAdmin(admin.ModelAdmin):
         'stdout_link',
         'stderr_link',
         'duration_str',
+        'hostname',
     )
     date_hierarchy = 'run_start_datetime'
     fieldsets = (
@@ -432,6 +483,7 @@ class LogAdmin(admin.ModelAdmin):
                 'run_end_datetime',
                 'duration_seconds',
                 'duration_str',
+                'hostname',
             )
         }),
         ('Output', {
@@ -444,6 +496,20 @@ class LogAdmin(admin.ModelAdmin):
         }),
     )
     
+    def queryset(self, *args, **kwargs):
+        qs = super(LogAdmin, self).queryset(*args, **kwargs)
+        qs = qs.only(
+            'id',
+            'run_start_datetime',
+            'run_end_datetime',
+            'duration_seconds',
+            'success',
+            'on_time',
+        )
+        if ApproxCountQuerySet is not None:
+            qs = qs._clone(klass=ApproxCountQuerySet)
+        return qs
+    
     def stdout_link(self, obj):
         url = reverse("admin:chroniker_log_change", args=(obj.id,)) + 'stdout/'
         return '<a href="%s">Download</a>' % (url,)
@@ -491,9 +557,10 @@ class LogAdmin(admin.ModelAdmin):
     job_name.short_description = _(u'Name')
 
     def job_success(self, obj):
-            return obj.success
+        return obj.success
     job_success.short_description = _(u'OK')
     job_success.boolean = True
+    job_success.admin_order_field = 'success'
     
     def has_add_permission(self, request):
         return False
@@ -576,8 +643,8 @@ class MonitorAdmin(admin.ModelAdmin):
     def action_buttons(self, obj):
         buttons = []
         url = '%d/run/?inline=1' % obj.id
-        buttons.append('<a href="%s"><input type="button" value="Check now" /></a>' % url)
-        buttons.append('<a href="/admin/chroniker/job/%i/" target="_blank"><input type="button" value="Edit" /></a>' % (obj.id,))
+        buttons.append('<a href="%s" class="button">Check now</a>' % url)
+        buttons.append('<a href="/admin/chroniker/job/%i/" target="_blank" class="button">Edit</a>' % (obj.id,))
         return ' '.join(buttons)
     action_buttons.allow_tags = True
     action_buttons.short_description = 'Actions'
