@@ -24,7 +24,7 @@ from django.conf import settings
 from django.contrib.sites.models import Site
 from django.core.mail import send_mail
 from django.core.management import call_command
-from django.db import models, connection
+from django.db import models, connection, transaction
 from django.db.models import Q
 from django.template import loader, Context, Template
 from django.utils import timezone
@@ -32,6 +32,7 @@ from django.utils.encoding import smart_str
 from django.utils.safestring import mark_safe
 from django.utils.timesince import timeuntil
 from django.utils.translation import ungettext, ugettext, ugettext_lazy as _
+from django.contrib.sites.models import get_current_site
 
 import chroniker.constants as const
 import chroniker.utils
@@ -41,6 +42,25 @@ logger = logging.getLogger('chroniker.models')
 _state = {} # {thread_ident:job_id}
 _state_heartbeat = {} # {thread_ident:heartbeat thread object}
 
+def order_by_dependencies(j1, j2):
+    """
+    Compares jobs so that jobs that are depended upon by another
+    process are ordered first.
+    """
+    # Return negative if x<y, zero if x==y, positive if x>y.
+    # If j1 depends on the completion of j2 then order j2 first.
+    if j1.dependencies.filter(dependee=j2).count():
+        ret = 1
+    elif j1.dependents.filter(dependent=j2).count():
+        ret = -1
+    elif j2.dependencies.filter(dependee=j1).count():
+        ret = -1
+    elif j2.dependents.filter(dependent=j1).count():
+        ret = 1
+    else:
+        ret = 0
+    return ret
+
 def get_current_job():
     """
     Retrieves the job associated with the current thread.
@@ -79,108 +99,6 @@ def set_current_heartbeat(obj):
     if thread_ident not in _state_heartbeat:
         _state_heartbeat[thread_ident] = obj
 
-class JobManager(models.Manager):
-    
-    def due(self, job=None):
-        """
-        Returns a ``QuerySet`` of all jobs waiting to be run.  NOTE: this may
-        return ``Job``s that are still currently running; it is your
-        responsibility to call ``Job.check_is_running()`` to determine whether
-        or not the ``Job`` actually needs to be run.
-        """
-        
-        # Lock the Job record if possible with the backend.
-        # https://docs.djangoproject.com/en/dev/ref/models/querysets/#select-for-update
-        # Note, select_for_update() may not be supported, such as with MySQL.
-        # Those backends will need to use an explicit backend-specific locking
-        # mechanism.
-        if settings.CHRONIKER_SELECT_FOR_UPDATE:
-            q = self.select_for_update(nowait=False)
-        else:
-            q = self.all()
-        q = q.filter(Q(next_run__lte=timezone.now()) | Q(force_run=True))
-        q = q.filter(
-            Q(hostname__isnull=True) | \
-            Q(hostname='') | \
-            Q(hostname=socket.gethostname()))
-        q = q.filter(enabled=True)
-        q = q.filter(is_running=False)
-        if job is not None:
-            if isinstance(job, int):
-                job = job.id
-            q = q.filter(id=job.id)
-        return q
-
-    def due_with_met_dependencies(self, jobs=[]):
-        """
-        Iterates over the results of due(), ignoring jobs
-        that are dependent on another job that is also due.
-        """
-        
-#        def cmp_deps(j1, j2):
-#            a = j1.dependencies.filter(dependee=j2).count()
-#            b = j2.dependencies.filter(dependee=j1).count()
-#            if a and not b:
-#                return +1
-#            elif not a and b:
-#                return -1
-#            return 0
-#        
-#        return sorted(self.due(), cmp=cmp_deps)
-        
-        # Fixes the "Lost connection to MySQL server during query" error when
-        # called from cron command?
-        connection.close()
-        
-        skipped_job_ids = set()
-        for job in self.due():
-            if jobs and job.id not in jobs:
-                print 'Skipping job %i (%s) because jobs are limited to %s.' % (job.id, job, ', '.join(map(str, jobs)))
-                skipped_job_ids.add(job.id)
-                continue
-            
-            deps = job.dependencies.all()
-            valid = True
-            
-            if job.check_is_running():
-                print 'Skipping job %i (%s) which is already running.' % (job.id, job)
-                #skipped_job_ids.add(job.id)
-                continue
-            
-            failed_dep = None
-            for dep in deps:
-                if dep.dependee.id in skipped_job_ids:
-                    continue
-                #elif dep.wait_for_completion and dep.dependee.is_due():
-                elif not dep.criteria_met():
-                    valid = False
-                    failed_dep = dep
-                    break
-                
-            if not valid:
-                print 'Skipping job %i (%s) which is dependent on a due job %i (%s).' \
-                    % (job.id, job, failed_dep.dependee.id, failed_dep.dependee)
-                skipped_job_ids.add(job.id)
-                continue
-            
-            #TODO:remove? redundant?
-            if not job.dependencies_met():
-                print 'Skipping job %i (%s) which has unmet dependencies.' % (job.id, job)
-                skipped_job_ids.add(job.id)
-                continue
-            
-            yield job
-
-    def stale(self):
-        q = self.filter(is_running=True)
-        q = q.filter(
-            Q(last_heartbeat__isnull=True) | 
-            Q(last_heartbeat__lt=timezone.now() - timedelta(minutes=5)))
-        return q
-    
-    def all_running(self):
-        return self.filter(is_running=True)
-
 class JobHeartbeatThread(threading.Thread):
     """
     A very simple thread that updates a temporary "lock" file every second.
@@ -208,6 +126,7 @@ class JobHeartbeatThread(threading.Thread):
         """
         Do not call this directly; call ``start()`` instead.
         """
+        check_freq_secs = 5
         while not self.halt:
             #print 'Heartbeat check...'
             self.lock_file.seek(0)
@@ -217,21 +136,23 @@ class JobHeartbeatThread(threading.Thread):
             # Check job status and save heartbeat timestamp.
             self.lock.acquire()
             Job.objects.update()
-            job = Job.objects.get(id=self.job_id)
-            job.last_heartbeat = timezone.now()
+            job = Job.objects.only('id', 'force_stop').get(id=self.job_id)
             force_stop = job.force_stop
-            job.force_stop = False
-            job.force_run = False
-            job.save()
+            Job.objects.filter(id=self.job_id).update(
+                last_heartbeat = timezone.now(),
+                force_stop = False,
+                force_run = False,
+            )
             self.lock.release()
             
-            # If we noticed we're being forced to stop, then interrupt main.
+            # If we noticed we're being forced to stop, then interrupt
+            # the entire process.
             if force_stop:
                 self.halt = True
                 thread.interrupt_main()
                 return
             
-            time.sleep(10)
+            time.sleep(check_freq_secs)
             
         set_current_heartbeat(None)
     
@@ -241,19 +162,18 @@ class JobHeartbeatThread(threading.Thread):
         """
         self.halt = True
         while self.is_alive():
-            #print 'Waiting for heartbeat to stop...'
             time.sleep(.1)
         self.lock_file.close()
         
-    def update_progress(self, total_parts, total_parts_complete):
-        print 'Updating progress:', total_parts, total_parts_complete
-        self.lock.acquire()
-        Job.objects.update()
-        job = Job.objects.get(id=self.job_id)
-        job.total_parts = total_parts
-        job.total_parts_complete = total_parts_complete
-        job.save()
-        self.lock.release()
+    def update_progress(self, total_parts, total_parts_complete, lock=True):
+        if lock:
+            self.lock.acquire()
+        Job.objects.filter(id=self.job_id).update(
+            total_parts = total_parts,
+            total_parts_complete = total_parts_complete,
+        )
+        if lock:
+            self.lock.release()
 
 class JobDependency(models.Model):
     """
@@ -285,31 +205,33 @@ class JobDependency(models.Model):
         help_text='If checked, the dependent job will not run until ' + \
             'the dependee job has a next_run greater than its next_run.')
     
-    def criteria_met(self):
-        if self.wait_for_completion and self.dependee.is_running:
+    def criteria_met(self, running_ids=None):
+        if running_ids is None:
+            running_ids = set()
+        if self.wait_for_completion and (self.dependee.is_running or self.dependee.id in running_ids):
             # Don't run until our dependency completes.
-            print '"%s": Dependee "%s" is still running.' \
-                % (self.dependent.name, self.dependee.name,)
+#            print '"%s": Dependee "%s" is still running.' \
+#                % (self.dependent.name, self.dependee.name,)
             return False
         if self.wait_for_success and not self.dependee.last_run_successful:
             # Don't run until our dependency completes successfully.
-            print '"%s": Dependee "%s" failed its last run.' \
-                % (self.dependent.name, self.dependee.name,)
+#            print '"%s": Dependee "%s" failed its last run.' \
+#                % (self.dependent.name, self.dependee.name,)
             return False
         if self.wait_for_next_run:
             # Don't run until our dependency is scheduled until after
             # our next run.
             if not self.dependent.next_run:
-                print '"%s": Our next scheduled run has not been set.' \
-                    % (self.dependent.name,)
+#                print '"%s": Our next scheduled run has not been set.' \
+#                    % (self.dependent.name,)
                 return False
             if not self.dependee.next_run:
-                print '"%s": Dependee "%s" has not been scheduled to run.' \
-                    % (self.dependent.name, self.dependee.name,)
+#                print '"%s": Dependee "%s" has not been scheduled to run.' \
+#                    % (self.dependent.name, self.dependee.name,)
                 return False
             if self.dependee.next_run < self.dependent.next_run:
-                print '"%s": Dependee "%s" has not yet run before us.' \
-                    % (self.dependent.name, self.dependee.name,)
+#                print '"%s": Dependee "%s" has not yet run before us.' \
+#                    % (self.dependent.name, self.dependee.name,)
                 return False
         return True
     criteria_met.boolean = True
@@ -319,12 +241,170 @@ class JobDependency(models.Model):
         unique_together = (
             ('dependent', 'dependee'),
         )
+        
+    def __unicode__(self):
+        return unicode(self.dependent) + ' -> ' + unicode(self.dependee)
+
+class JobManager(models.Manager):
+    
+    def due(self, job=None, check_running=True):
+        """
+        Returns a ``QuerySet`` of all jobs waiting to be run.  NOTE: this may
+        return ``Job``s that are still currently running; it is your
+        responsibility to call ``Job.check_is_running()`` to determine whether
+        or not the ``Job`` actually needs to be run.
+        """
+        
+        # Lock the Job record if possible with the backend.
+        # https://docs.djangoproject.com/en/dev/ref/models/querysets/#select-for-update
+        # Note, select_for_update() may not be supported, such as with MySQL.
+        # Those backends will need to use an explicit backend-specific locking
+        # mechanism.
+        if settings.CHRONIKER_SELECT_FOR_UPDATE:
+            q = self.select_for_update(nowait=False)
+        else:
+            q = self.all()
+        q = q.filter(Q(next_run__lte=timezone.now()) | Q(force_run=True))
+        q = q.filter(
+            Q(hostname__isnull=True) | \
+            Q(hostname='') | \
+            Q(hostname=socket.gethostname()))
+        q = q.filter(enabled=True)
+        if check_running:
+            q = q.filter(is_running=False)
+        if job is not None:
+            if isinstance(job, int):
+                job = job.id
+            q = q.filter(id=job.id)
+        return q
+
+    def due_with_met_dependencies(self, jobs=[]):
+        """
+        Iterates over the results of due(), ignoring jobs
+        that are dependent on another job that is also due.
+        """
+        
+        # Fixes the "Lost connection to MySQL server during query" error when
+        # called from cron command?
+        connection.close()
+        
+        skipped_job_ids = set()
+        for job in self.due():
+            if jobs and job.id not in jobs:
+                #print 'Skipping job %i (%s) because jobs are limited to %s.' % (job.id, job, ', '.join(map(str, jobs)))
+                skipped_job_ids.add(job.id)
+                continue
+            
+            deps = job.dependencies.all()
+            valid = True
+            
+            if job.check_is_running():
+                #print 'Skipping job %i (%s) which is already running.' % (job.id, job)
+                continue
+            
+            failed_dep = None
+            for dep in deps:
+                if dep.dependee.id in skipped_job_ids:
+                    continue
+                #elif dep.wait_for_completion and dep.dependee.is_due():
+                elif not dep.criteria_met():
+                    valid = False
+                    failed_dep = dep
+                    break
+                
+            if not valid:
+                #print 'Skipping job %i (%s) which is dependent on a due job %i (%s).' \
+                #    % (job.id, job, failed_dep.dependee.id, failed_dep.dependee)
+                skipped_job_ids.add(job.id)
+                continue
+            
+            #TODO:remove? redundant?
+            if not job.dependencies_met():
+                #print 'Skipping job %i (%s) which has unmet dependencies.' % (job.id, job)
+                skipped_job_ids.add(job.id)
+                continue
+            
+            yield job
+
+    def due_with_met_dependencies_ordered(self, jobs=[]):
+        """
+        Returns a list of jobs sorted by dependency, with dependents after
+        all their dependees.
+        """
+        lst = list(self.due_with_met_dependencies(jobs=jobs))
+        lst.sort(cmp=order_by_dependencies)
+        return lst
+
+    def stale(self):
+        q = self.filter(is_running=True)
+        q = q.filter(
+            Q(last_heartbeat__isnull=True) | 
+            Q(last_heartbeat__lt=timezone.now() - timedelta(minutes=settings.CHRONIKER_STALE_MINUTES)))
+        return q
+    
+    def all_running(self):
+        return self.filter(is_running=True)
+    
+    @transaction.commit_manually
+    def end_all_stale(self):
+        """
+        Marks as complete but failed, and attempts to kill the process
+        if still running, any job that's failed to report its status within
+        the alotted threshold.
+        """
+        try:
+            q = self.stale()
+            total = q.count()
+            print '%i total stale jobs.' % (total,)
+            for job in q.iterator():
+                print 'Checking stale job %i <%s>...' % (job.id, job,)
+                
+                # If we know the PID and it's running locally, and the process
+                # appears inactive, then attempt to forcibly kill the job.
+                if job.current_pid and job.current_hostname \
+                and job.current_hostname == socket.gethostname():
+                    if chroniker.utils.pid_exists(job.current_pid):
+#                        cpu_usage = chroniker.utils.get_cpu_usage(job.current_pid)
+#                        if cpu_usage:
+#                            print 'Process with PID %s is still consuming CPU so keeping alive for now.' % (job.current_pid,)
+#                            continue
+#                        else:
+                        print 'Killing process %s...' % (job.current_pid,)
+                        chroniker.utils.kill_process(job.current_pid)
+                        #TODO:record log entry
+                    else:
+                        print 'Process with PID %s is not running.' % (job.current_pid,)
+                else:
+                    print 'Process with PID %s is not elligible for killing.' % (job.current_pid,)
+                
+                job.is_running = False
+                job.last_run_successful = False
+                job.save()
+                transaction.commit()
+                
+                Log.objects.create(
+                    job = job,
+                    run_start_datetime = job.last_run_start_timestamp or timezone.now(),
+                    run_end_datetime = timezone.now(),
+                    hostname = socket.gethostname(),
+                    stdout = '',
+                    stderr = 'Job became stale and was marked as terminated.',
+                    success = False,
+                )
+                transaction.commit()
+        except:
+            transaction.rollback()
+            raise
+        else:
+            transaction.commit()
 
 class Job(models.Model):
     """
     A recurring ``django-admin`` command to be run.
     """
     
+    objects = JobManager()
+    
     name = models.CharField(
         _("name"),
         max_length=200)
@@ -358,7 +438,8 @@ class Job(models.Model):
     
     enabled = models.BooleanField(
         default=True,
-        help_text=_('If checked this job will run.'))
+        help_text=_('''If checked, this job will be run automatically according
+            to the frequency options.'''))
     
     next_run = models.DateTimeField(
         _("next run"),
@@ -419,11 +500,18 @@ class Job(models.Model):
     
     force_run = models.BooleanField(
         default=False,
-        help_text=_("If checked then this job will be run immediately."))
+        help_text=_("If checked, then this job will be run immediately."))
     
     force_stop = models.BooleanField(
         default=False,
-        help_text=_("If checked and running then this job will be stopped."))
+        help_text=_("If checked, and running then this job will be stopped."))
+    
+    timeout_seconds = models.PositiveIntegerField(
+        default=0,
+        blank=False,
+        null=False,
+        help_text=_('''When non-zero, the job will be forcibly killed if running
+            for more than this amount of time.'''))
     
     hostname = models.CharField(
         max_length=700,
@@ -443,6 +531,14 @@ class Job(models.Model):
         editable=False,
         help_text=_('The name of the host currently running the job.'))
     
+    current_pid = models.CharField(
+        max_length=50,
+        blank=True,
+        null=True,
+        editable=False,
+        db_index=True,
+        help_text=_('The ID of the process currently running the job.'))
+    
     total_parts_complete = models.PositiveIntegerField(
         default=0,
         editable=False,
@@ -481,14 +577,22 @@ class Job(models.Model):
         null=True,
         #verbose_name='records',
         editable=False,
-        help_text=_('The number of records that needs attention.'))
+        help_text=_('The number of records that need attention.'))
     
     maximum_log_entries = models.PositiveIntegerField(
         default=1000,
         help_text='The maximum number of most recent log entries to keep.' + \
             '<br/>A value of 0 keeps all log entries.')
     
-    objects = JobManager()
+    log_stdout = models.BooleanField(
+        default=True,
+        help_text=_('''If checked, all characters printed to stdout will be
+            saved in a log record.'''))
+    
+    log_stderr = models.BooleanField(
+        default=True,
+        help_text=_('''If checked, all characters printed to stderr will be
+            saved in a log record.'''))
     
     class Meta:
         ordering = (
@@ -499,7 +603,7 @@ class Job(models.Model):
     def __unicode__(self):
         if not self.enabled:
             return _(u"%(id)s - %(name)s - disabled") % {'name': self.name, 'id':self.id}
-        return u"%s - %s" % (self.name, self.timeuntil)
+        return u"%i - %s - %s" % (self.id, self.name, self.timeuntil)
     
     @property
     def monitor_url_rendered(self):
@@ -539,6 +643,36 @@ class Job(models.Model):
         return '%.0f%%' % (progress,)
     progress_percent_str.short_description = 'Progress'
     
+    def get_chained_jobs(self):
+        """
+        Returns a list of jobs that depend on this job.
+        Retrieves jobs recursively, stopping if it detects cycles.
+        """
+        priors = set([self.id])
+        pending = list(self.dependents.all().filter(dependent__enabled=True, wait_for_completion=True).values_list('dependent_id', flat=True))
+        chained = set()
+        while pending:
+            job_id = pending.pop(0)
+            if job_id in priors:
+                continue
+            priors.add(job_id)
+            job = Job.objects.only('id').get(id=job_id)
+            chained.add(job)
+            pending.extend(job.dependents.all().filter(dependent__enabled=True, wait_for_completion=True).values_list('dependent_id', flat=True))
+        return chained
+    
+    def get_run_length_estimate(self, samples=20):
+        """
+        Returns the average run length in seconds.
+        """
+        q = sorted(list(self.logs.all().values_list('duration_seconds', flat=True).order_by('-run_end_datetime')[:samples]))
+        if len(q) >= 3:
+            # Drop the upper and lower extremes.
+            q = q[1:-1]
+        if not q:
+            return
+        return int(round(sum(q)/float(len(q))))
+    
     @property
     def estimated_seconds_to_completion(self):
         """
@@ -553,6 +687,8 @@ class Job(models.Model):
         if not self.last_run_start_timestamp:
             return
         td = timezone.now() - self.last_run_start_timestamp
+        if not progress_ratio:
+            return
         total_sec = td.seconds/progress_ratio
         remaining_sec = total_sec - td.seconds
         return remaining_sec
@@ -583,47 +719,46 @@ class Job(models.Model):
                 logger.debug("Updating 'next_run")
                 next_run = self.next_run or timezone.now()
                 
-                #TODO:why can't rrule consistently handle timezone-aware datetimes?!?!?!
                 tz = timezone.get_default_timezone()
                 try:
                     self.next_run = self.rrule.after(timezone.make_aware(next_run, tz))
+                except ValueError, e:
+                    self.next_run = timezone.make_aware(
+                        self.rrule.after(timezone.make_naive(next_run, tz)),
+                        tz)
                 except TypeError, e:
                     self.next_run = timezone.make_aware(
                         self.rrule.after(timezone.make_naive(next_run, tz)),
                         tz)
         
-        #old = None
-        #if self.id:
-        #    old = Job.objects.get(id=self.id)
-        
         if not self.is_running:
             self.current_hostname = None
+            self.current_pid = None
         
         super(Job, self).save(*args, **kwargs)
         
-        #if old:
-        #    print 'is_running_changed:', old.is_running != self.is_running, old.is_running, '->', self.is_running
-        
         # Delete expired logs.
         if self.maximum_log_entries:
             log_q = self.logs.all().order_by('-run_start_datetime')
             for o in log_q[self.maximum_log_entries:]:
                 o.delete()
 
-    def dependencies_met(self):
+    def dependencies_met(self, running_ids=None):
         """
         Returns true if all dependency scheduling criteria have been met.
         Returns false otherwise.
         """
         for dep in self.dependencies.all():
-            if not dep.criteria_met():
+            if not dep.criteria_met(running_ids=running_ids):
                 return False
         return True
 
     def is_fresh(self):
+        lookback_minutes = timedelta(minutes=settings.CHRONIKER_STALE_MINUTES)
+        threshold = timezone.now() - lookback_minutes
         return not self.is_running or (
             self.is_running and self.last_heartbeat
-            and self.last_heartbeat >= timezone.now() - timedelta(minutes=5)
+            and self.last_heartbeat >= threshold
         )
     is_fresh.boolean = True
 
@@ -747,7 +882,7 @@ class Job(models.Model):
                 args.append(arg)
         return (args, options)
     
-    def is_due(self):
+    def is_due(self, **kwargs):
         """
         >>> job = Job(next_run=timezone.now())
         >>> job.is_due()
@@ -765,30 +900,37 @@ class Job(models.Model):
         >>> job.is_due()
         False
         """
-#        reqs =  (
-#            self.next_run <= timezone.now()
-#            and self.enabled
-#            and self.check_is_running() == False
-#        )
-#        return (reqs or self.force_run)
-        q = type(self).objects.due(self)
-        return bool(q.count())
+        q = type(self).objects.due(job=self, **kwargs)
+        return q.exists()
+    is_due.boolean = True
     
-    def run(self, *args, **kwargs):
+    def is_due_with_dependencies_met(self, running_ids=None):
+        """
+        Return true if job is scheduled to run and all dependencies
+        are satisified.
+        """
+        return self.is_due() and self.dependencies_met(running_ids=running_ids)
+    
+    def run(self, check_running=True, *args, **kwargs):
         """
         Runs this ``Job``.  A ``Log`` will be created if there is any output
         from either stdout or stderr.
         
         Returns ``True`` if the ``Job`` ran, ``False`` otherwise.
+        
+        The parameter check_running is sometimes given as False when
+        the job was just started and we want to do a last minute check for
+        dueness and don't want our current run status to give an incorrect
+        reading.
         """
         if self.enabled:
             if not self.dependencies_met():
                 # Note, this will cause the job to be re-checked
                 # the next time cron runs.
                 print 'Job "%s" has unmet dependencies. Aborting run.' % (self.name,)
-            elif self.check_is_running():
+            elif check_running and self.check_is_running():
                 print 'Job "%s" already running. Aborting run.' % (self.name,)
-            elif not self.is_due():
+            elif not self.is_due(check_running=check_running):
                 print 'Job "%s" not due. Aborting run.' % (self.name,)
             else:
                 #call_command('run_job', str(self.pk)) # Calls handle_run().
@@ -798,7 +940,7 @@ class Job(models.Model):
             print 'Job disabled. Aborting run.'
         return False
     
-    def handle_run(self, update_heartbeat=True, *args, **kwargs):
+    def handle_run(self, update_heartbeat=True, stdout_queue=None, stderr_queue=None, *args, **kwargs):
         """
         This method implements the code to actually run a ``Job``.  This is
         meant to be run, primarily, by the `run_job` management command as a
@@ -810,10 +952,19 @@ class Job(models.Model):
         lock = threading.Lock()
         run_start_datetime = timezone.now()
         last_run_successful = False
-        stdout = chroniker.utils.TeeFile(sys.stdout, auto_flush=True)
-        stderr = chroniker.utils.TeeFile(sys.stderr, auto_flush=True)
-        t0 = time.time()
         
+        stdout = sys.stdout
+        stderr = sys.stderr
+        if self.log_stdout:
+            stdout = chroniker.utils.TeeFile(
+                sys.stdout,
+                auto_flush=True,
+                queue=stdout_queue)
+        if self.log_stderr:
+            stderr = chroniker.utils.TeeFile(
+                sys.stderr,
+                auto_flush=True,
+                queue=stderr_queue)
         try:
     
             # Redirect output so that we can log it if there is any
@@ -830,16 +981,26 @@ class Job(models.Model):
 
             try:
                 lock.acquire()
-                Job.objects.update()
-                job = Job.objects.get(id=self.id)
-                job.is_running = True
-                job.last_run_start_timestamp = timezone.now()
-                job.current_hostname = socket.gethostname()
-                job.total_parts = 0
-                job.total_parts_complete = 0
-                if heartbeat:
-                    job.lock_file = heartbeat.lock_file.name
-                job.save()
+#                Job.objects.update()
+#                job = Job.objects.get(id=self.id)
+#                job.is_running = True
+#                job.last_run_start_timestamp = timezone.now()
+#                job.current_hostname = socket.gethostname()
+#                job.current_pid = str(os.getpid())
+#                job.total_parts = 0
+#                job.total_parts_complete = 0
+#                if heartbeat:
+#                    job.lock_file = heartbeat.lock_file.name
+#                job.save()
+                Job.objects.filter(id=self.id).update(
+                    is_running = True,
+                    last_run_start_timestamp = timezone.now(),
+                    current_hostname = socket.gethostname(),
+                    current_pid = str(os.getpid()),
+                    total_parts = 0,
+                    total_parts_complete = 0,
+                    lock_file = heartbeat.lock_file.name if heartbeat and heartbeat.lock_file else '',
+                )
             except Exception, e:
                 # The command failed to run; log the exception
                 t = loader.get_template('chroniker/error_message.txt')
@@ -852,7 +1013,6 @@ class Job(models.Model):
             finally:
                 lock.release()
             
-            t0 = time.time()
             if heartbeat:
                 heartbeat.start()
             success = True
@@ -896,17 +1056,27 @@ class Job(models.Model):
             try:
                 lock.acquire()
                 Job.objects.update()
-                job = Job.objects.get(id=self.id)
-                job.is_running = False
-                job.lock_file = ""
-                job.last_run = run_start_datetime
-                job.force_run = False
-                job.next_run = next_run
-                job.last_run_successful = last_run_successful
-                # Ensure we report 100% progress if everything ran successfully.
-                if job.last_run_successful and job.total_parts is not None:
-                    job.total_parts_complete = job.total_parts
-                job.save()
+#                job = Job.objects.get(id=self.id)
+#                job.is_running = False
+#                job.lock_file = ""
+#                job.last_run = run_start_datetime
+#                job.force_run = False
+#                job.next_run = next_run
+#                job.last_run_successful = last_run_successful
+#                # Ensure we report 100% progress if everything ran successfully.
+#                if job.last_run_successful and job.total_parts is not None:
+#                    job.total_parts_complete = job.total_parts
+#                job.save()
+                job = Job.objects.only('id', 'total_parts', 'last_run_successful').get(id=self.id)
+                Job.objects.filter(id=self.id).update(
+                    is_running = False,
+                    lock_file = '',
+                    last_run = run_start_datetime,
+                    force_run = False,
+                    next_run = next_run,
+                    last_run_successful = last_run_successful,
+                    total_parts_complete = (job.last_run_successful and job.total_parts) or 0,
+                )
             except Exception, e:
                 # The command failed to run; log the exception
                 t = loader.get_template('chroniker/error_message.txt')
@@ -927,23 +1097,33 @@ class Job(models.Model):
             
             # Record run log.
             print 'Recording log...'
-            stdout = stdout.getvalue()
-            if isinstance(stdout, unicode):
-                stdout = stdout.encode('utf-8', 'replace')
-            else:
-                stdout = unicode(stdout, 'utf-8', 'replace')
-            stderr = stderr.getvalue()
-            if isinstance(stderr, unicode):
-                stderr = stderr.encode('utf-8', 'replace')
-            else:
-                stderr = unicode(stderr, 'utf-8', 'replace')
+            
+            stdout_str = ''
+            if self.log_stdout:
+                stdout_str = stdout.getvalue()
+                if isinstance(stdout_str, unicode):
+                    stdout_str = stdout_str.encode('utf-8', 'replace')
+                else:
+                    stdout_str = unicode(stdout_str, 'utf-8', 'replace')
+            
+            stderr_str = ''
+            if self.log_stderr:
+                stderr_str = stderr.getvalue()
+                if isinstance(stderr_str, unicode):
+                    stderr_str = stderr_str.encode('utf-8', 'replace')
+                else:
+                    stderr_str = unicode(stderr_str, 'utf-8', 'replace')
+            
+            run_end_datetime = timezone.now()
+            duration_seconds = (run_end_datetime - run_start_datetime).seconds
             log = Log.objects.create(
                 job = self,
                 run_start_datetime = run_start_datetime,
-                run_end_datetime = timezone.now(),
-                duration_seconds = time.time() - t0,
-                stdout = stdout,
-                stderr = stderr,
+                run_end_datetime = run_end_datetime,
+                duration_seconds = duration_seconds,
+                hostname=socket.gethostname(),
+                stdout = stdout_str,
+                stderr = stderr_str,
                 success = last_run_successful,
             )
             
@@ -1007,43 +1187,73 @@ class Job(models.Model):
         if heartbeat:
             return heartbeat.update_progress(*args, **kwargs)
         else:
-            print 'Unable to update progress. No heartbeat found.'
+            #print 'Unable to update progress. No heartbeat found.'
+            pass
 
 class Log(models.Model):
     """
     A record of stdout and stderr of a ``Job``.
     """
+    
     job = models.ForeignKey(Job, related_name='logs')
+    
     run_start_datetime = models.DateTimeField(
         editable=False,
         db_index=True,
         default=timezone.now,
         blank=False,
         null=False)
+    
     run_end_datetime = models.DateTimeField(
         editable=False,
         db_index=True,
         blank=True,
         null=True)
+    
     duration_seconds = models.PositiveIntegerField(
         editable=False,
         db_index=True,
         verbose_name='duration (total seconds)',
         blank=True,
         null=True)
+    
     stdout = models.TextField(blank=True)
+    
     stderr = models.TextField(blank=True)
+    
+    hostname = models.CharField(
+        max_length=700,
+        blank=True,
+        null=True,
+        editable=False,
+        help_text=_('The hostname this job was executed on.'))
+    
     success = models.BooleanField(
         default=True,
         db_index=True,
         editable=False)
         
+    on_time = models.BooleanField(
+        default=True,
+        db_index=True,
+        editable=False,
+        help_text=_('''If true, indicates job completed of its own accord.
+            If false, the job exceeded a timeout threshold and was forcibly
+            killed.'''))
+    
     class Meta:
         ordering = ('-run_start_datetime',)
     
     def __unicode__(self):
         return u"%s - %s" % (self.job.name, self.run_start_datetime)
     
+    def save(self, *args, **kwargs):
+        if self.run_start_datetime and self.run_end_datetime:
+            assert self.run_start_datetime <= self.run_end_datetime, 'Job must start before it ends.'
+            time_diff = (self.run_end_datetime - self.run_start_datetime)
+            self.duration_seconds = time_diff.seconds
+        super(Log, self).save(*args, **kwargs)
+    
     def duration_str(self):
         from datetime import datetime, timedelta
         sec = timedelta(seconds=self.duration_seconds)
@@ -1084,6 +1294,23 @@ class Log(models.Model):
         else:
             body = "Ouput:\n%s\nError output:\n%s" % (self.stdout, self.stderr)
         
+        #site = get_current_site()
+        site = Site.objects.get_current()
+        base_url = None
+        if hasattr(settings, 'BASE_SECURE_URL'):
+            base_url = settings.BASE_SECURE_URL
+        elif hasattr(settings, 'BASE_URL'):
+            base_url = settings.BASE_URL
+        elif site:
+            if domain.startswith('http'):
+                base_url = site.domain
+            else:
+                base_url = 'http://' + site.domain
+        
+        if base_url:
+            admin_link = settings.BASE_SECURE_URL + chroniker.utils.get_admin_change_url(self.job)
+            body = 'To manage this job please visit: ' + admin_link + '\n\n' + body
+        
         send_mail(
             from_email = '"%s" <%s>' % (
                 settings.CHRONIKER_EMAIL_SENDER,
