@@ -14,4 +14,14 @@ class Sleeper(BaseCommand):
         time.sleep(target_time)
         
         end_time = time.time()
-        print "Job ran for %f seconds" % (end_time-start_time)
\ No newline at end of file
+        print "Job ran for %f seconds" % (end_time-start_time)
+
+class InfiniteWaiter(BaseCommand):
+    args = ''
+    help = 'A simple command that waits indefinitely.'
+    
+    def handle(self, *args, **options):
+        while 1:
+            time.sleep(1)
+            print 'Waiting...'
+            
\ No newline at end of file
